unit ABitmap;
// FITS Bildanalyse
// O. Nickel, Mainz 2005

interface

uses
  Windows, Graphics, Messages, Dialogs, SysUtils, Forms, classes,
  Gmlib,astro1{,dateutils};

  { IMAGE BUFFERS }
  {---------------}
const
  ImageCacheEnabled : boolean=false;
  ErrorReporting: boolean=false;
  Maxheaderlines=144;
type
  wordarray = array[1..10000000] of word;
  bytearray = array[1..10000000] of byte;
  intarray  = array[1..10000000] of smallint;
  longarray = array[1..10000000] of longint;
  Tfitsline = array[1..80] of char;
  Tfitsheader = array[1..Maxheaderlines] of Tfitsline;
  Tfitskey = string;

  TPalette256 = packed record
    palVersion: Word;
    palNumEntries: Word;
    palPalEntry: array[0..255] of TPaletteEntry;
  end;
  TBitmapInfo256 = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array[0..255] of TRGBQuad;
  end;

  Timgbuffer = class(Tobject)
    bufferpointer : Pointer; (*pointer to matrix data*)
    fitsheader: Tfitsheader;
    headerlines: integer;
    JDstart: double; (*Start time Jul.Datum*)
    Datetime: string; // UT Date and time
    Origin: string[20];
    Exposure: single; //exposure time in minutes
    xsize,ysize : Integer; (*matrixsize*)
    Xpixelsize,Ypixelsize: single; //pixel size in microns
    Instrument: string[20];
    Observer: string[20];
    Xbinning,Ybinning: word; //binning (1..3)
    RA, DEC: single; //telescope RA,DEC
    Xscale,Yscale: Single; //scale in rad/pixel
    yori : boolean; { orientation: True=from top to bottom}
    bufferbyte : Integer;
               (* pixel value format: 1= 8bit unsigned 2= 16bit unsigned
               4= 32bit signed (longint)*)
    bufferMax : integer; (* Max. pixel value without scaling *)
    bufferMin : integer; // Min pixel value
    Offset1: double; //Fits BZERO
    ScaleFactor: double; { BSCALE, Optional scaling factor, default = 1.0 }
    Wordmode: boolean;
    imageblock: longint;
    filename: string;
    shortInfo: string;

  private
    procedure gettemp(var imgtemp:Timgbuffer);
  protected

  public
    // initialization of image buffer
    constructor create;
    destructor destroy; override;
    function init(xdim,ydim,imgbyte:integer):integer;// image buffer is initialized
             // xdim,ydim: matrix size, imgbyte: 1 for 8bit, 2 (or 0) for 16bit buffer
             // all pixel values are set to 0

    // pixel routines
    function getpixel(x,y:integer):integer; // returns value of matrix point (x,y)
                                            // (x=1..xdim, y=1..ydim)
    procedure putpixel(x,y:integer; value:integer); // sets pixel (x,y) to <value>
                                            // (x=1..xdim, y=1..ydim)
    procedure Getline(y:integer; var line);  // Gammalib:MoveMiline(...)
    procedure Putline(var line; y:integer);  // Gammalib:MovelinetoBuffer(...)
    procedure Getcolumn(x:integer; var col); // Gammalib:MoveMicolumn(...)
    procedure Putcolumn(var col; x:integer); // Gammalib:MoveColumntoBuffer(...)
    function getmaxpixel: integer; //returns maximum pixel value of image
    procedure getminmax;// integer array also

    // routines for whole image
    procedure swapbytes; //changes byte order of pixel values (word only)
    procedure flipvertical; // image is mirrored vertical
    procedure fliphorizontal; //image is mirrored horizontally
    procedure interpolate;//bilinear interpolation: Matrixsize m x n -> 2m x 2n
    function duplicate: Timgbuffer;  //returns duplicated image
    function AddImage(img:Timgbuffer):boolean;
             //adds image <img> to buffer. <img> must have equal size (xsize x ysize),
             // but can have different bits/pixel
    procedure ByteToWord; // changes byte matrix to word matrix
    procedure IntToWord(offset:integer);
              //changes signed (smallint) to word matrix
              //offset is added to pixel value, resulting value is truncated to word range
    procedure WordtoInt(offset:integer);
              //changes word to signed (smallint) matrix:
              //offset is subtracted from pixel value, resulting value is truncated to int range
    procedure FitsToLongint(offset:double);
              //subtracts fits offset from signed (longint) matrix:
    procedure UpdateFitsHeader;
    function Readfitsfile(fitsfilename:string):boolean;
    function Writefitsfile(fitsfilename:string):boolean;
    function ReadBMPfile(bmpfilename:string):boolean;

  published
  end;

  TcacheID = integer;
  TimageCache = class(Tobject)
    maxindex: integer; {maximum number of image buffers}
    maxmem: integer; {maximum size of cache memory}
    function PutImagetoCache(img:Timgbuffer):TcacheID; {puts image into cache, returned number = cacheID}
    function loadfromCache(cacheID:TcacheID; var img:Timgbuffer):integer;
    function DisposeCache(cacheID:TcacheID):integer; {disposes cached image with ID <cacheID>}

  private
    lastindex: integer;
    memcount: integer;
    cbuffer: array[1..2048] of Timgbuffer;
    ID: array[1..2048] of TcacheID;
    accesscount: array[1..2048] of integer;
    lastID: TcacheID;
    function getfreebuffer: integer;
  protected
    constructor create;
    destructor destroy; override;
  public
  published
  end;

//FITS
function setFitskey(var header:Tfitsheader; key:Tfitskey; value:string):boolean;

//Image cache
procedure StartImageCache(cacheMB:integer); //starts imagecache1 with size of cacheMB Megabytes
procedure StopImageCache;


  { BITMAPS }
  {---------}

const
  DCnumcolors : integer=256;
type

  TBitmapRGB = class(TBitmap)
  private
    { Private-Deklarationen }
  protected
    { Protected-Deklarationen }
  public
    { Public-Deklarationen }
    procedure LoadfromArray(arr:pointer; xsize,ysize:Integer);
  published
    { Published-Deklarationen }
  end;

  TBitmap256 = class(TBitmap)
  private
    { Private-Deklarationen }
  protected
    { Protected-Deklarationen }
  public
    destructor destroy; override;
    { Public-Deklarationen }
    function LoadfromArray(arr:pointer;xsize,ysize:Integer; Hpal:HPalette):boolean;
{ creates a 256 color bitmap from a byte array (pointer arr)
  with size xsize x ysize holding the bitmap pixels
  and a Palette handle HPal}
    procedure LoadfromBuffer(img: TImgBuffer;ilt,iut:integer;Hpal:Hpalette);
  published
    { Published-Deklarationen }
  end;

{Color Tables, Palette functions}
{-------------------------------}

function GetnumColors: integer;
         // returns the number of colors available on screen
function CreatePalette256(pal256: TPalette256):HPalette;
         // creates a windows palette from a 256 color palette
//function LoadPalette(p:integer):Hpalette;
         // obsolete
function GetColortableList(var CtableList,CtableFileList: Tstringlist):integer;
         // returns a list of color table names corresp. to a list
         // of filenames. Function value = number of table elements
function GetColortable(filename:string; var colortable:Tpalette256;
                        inverted:boolean): boolean;
         // returns a 256 color palette from a colortable file
         // if <inverted>=true, colors are inverted
         // format of colortable file:
         // ASCII-file, first line: GammaView colortable 256 R,G,B
         // second line: <name of colortable>
         // line 3..257 contain the color values for index 0..255
         // format <red>,<green>,<blue>
function GetGrayPalette(var gPal:TPalette256; inverted:boolean ):boolean;
         // returns a linear gray palette

var
  ImageCache1: TimageCache;
  CtableList1,CtableFileList1: Tstringlist;
  DC0: HDC;

implementation

{ IMAGE BUFFERS }
{---------------}

constructor Timgbuffer.Create;
begin
    inherited create;
    buffermax:=-1;
    buffermin:=-1;
    xsize:=0; ysize:=0;
    Yori:=True;
    Offset1:=0;  //BZERO
    Wordmode:=true;
    Imageblock:=0;
    filename:='';
    ScaleFactor :=1.0; { BSCALE, optional scaling factor
                          for quantitative evaluation,
                          default = 1.0 }
    ShortInfo:='';
    headerlines:=0;
    setfitskey(fitsheader,'SIMPLE','T');
    setfitskey(fitsheader,'BITPIX','16');
    setfitskey(fitsheader,'NAXIS','2');
    setfitskey(fitsheader,'BSCALE','1');
    setfitskey(fitsheader,'BZERO','0');
end;

destructor Timgbuffer.destroy;
begin
  if (bufferpointer<>nil) then freemem(bufferpointer);
  inherited destroy;
end;

function Timgbuffer.init(xdim,ydim,imgbyte: integer):integer;
var
  nbytes: integer;
begin
  init:=-1;
  if imgbyte=0 then nbytes:=2
             else nbytes:=imgbyte;
  if bufferpointer<>nil then freemem(bufferpointer); //reinit
  getmem(bufferpointer,xdim*ydim*nbytes);
  if bufferpointer=nil then exit;
  xsize:=xdim; ysize:=ydim;  bufferbyte:=imgbyte;
  fillchar(bufferpointer^,xdim*ydim*nbytes,0);
  init:=0;
end;

function Timgbuffer.getpixel(x,y:integer):integer;
var
  pb: ^byteArray;
  pw: ^wordArray;
  plong: ^longArray;
begin
  if xsize=0 then getpixel:=0
  else
  begin
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer;
           getpixel:=pw^[(y-1)*xsize+x]; //corr.: 18-jan-99
         end;
    1: begin
           pb:=bufferpointer;
           getpixel:=pb^[(y-1)*xsize+x]; //corr.: 18-jan-99
       end;
    4: begin
         plong:=bufferpointer;
         getpixel:=plong^[(y-1)*xsize+x];
       end;
    else getpixel:=0;
    end;
  end;
end;

procedure Timgbuffer.putpixel(x,y:integer; value:integer);
var
  pb: ^byteArray;
  pw: ^wordArray;
  plong: ^longArray;
begin
  if xsize=0 then exit
  else
  begin
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer;
           pw^[(y-1)*xsize+x]:=min(65535,value); //corr.: 18-jan-99
         end;
    1: begin
           pb:=bufferpointer;
           pb^[(y-1)*xsize+x]:=min(255,value); //corr.: 18-jan-99
       end;
    4: begin
         plong:=bufferpointer;
         plong^[(y-1)*xsize+x]:=value;
       end;
    end;
  end;
end;

function Timgbuffer.getmaxpixel:integer;// byte and word only
var
  pb: ^byteArray;
  pw: ^wordArray;
  plong: ^longArray;
  m,i,n: integer;
begin
  if xsize=0 then getmaxpixel:=0
  else
  begin
    m:=0;
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer;
           for i:=1 to xsize*ysize do m:=max(m,pw^[i]);
           getmaxpixel:=m;
         end;
    1: begin
           pb:=bufferpointer;
           for i:=1 to xsize*ysize do m:=max(m,pb^[i]);
           getmaxpixel:=m;
       end;
    4: begin
           plong:=bufferpointer;
           for i:=1 to xsize*ysize do m:=max(m,plong^[i]);
           getmaxpixel:=m;
       end;
    else getmaxpixel:=-1;
    end;
  end;
end;

procedure Timgbuffer.getminmax;// integer array also
var
  pb: ^byteArray;
  pw: ^wordArray;
  pi: ^IntArray;
  plong: ^longArray;
  m,i,n: integer;
begin
  if xsize=0 then exit;
  if Wordmode then
  begin
    pw:=bufferpointer;
    pb:=bufferpointer;
    m:=0; n:=65535;
    case bufferbyte of
    0,2: begin
           for i:=1 to xsize*ysize do
           begin
             m:=max(m,pw^[i]); n:=min(n,pw^[i]);
           end;
         end;
    1: begin
           for i:=1 to xsize*ysize do
           begin
             m:=max(m,pb^[i]);n:=min(n,pb^[i]);
           end;
       end;
    end;
  end else
  begin
    pi:=bufferpointer;plong:=bufferpointer;
    case bufferbyte of
    2: begin
         m:=pi^[1]; n:=m;
         for i:=1 to xsize*ysize do
         begin
              m:=max(m,pi^[i]);n:=min(n,pi^[i]);
         end;
       end;
    4: begin
         m:=plong^[1]; n:=m;
         for i:=1 to xsize*ysize do
         begin
              m:=max(m,plong^[i]);n:=min(n,plong^[i]);
         end;
       end;
    end;
  end;
  buffermax:=m; buffermin:=n;
end;

procedure Timgbuffer.Getline(y:integer; var line);
var
  pb: ^byteArray;
  pw: ^wordArray;
  plong: ^longArray;
  i : longint;
begin
  if xsize<=0 then exit
  else
  begin
    i:=(y-1)*xsize+1;
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer;
           move(pw^[i],line,xsize*2);
         end;
    1: begin
           pb:=bufferpointer;
           move(pb^[i],line,xsize);
      end;
    4: begin
           plong:=bufferpointer;
           move(plong^[i],line,xsize*4);
      end;
    end;
  end;

end;

procedure Timgbuffer.Putline(var line; y:integer);
var
  pb: ^byteArray;
  pw: ^wordArray;
  plong: ^longArray;
  i : longint;
begin
  if xsize=0 then exit
  else
  begin
    i:=(y-1)*xsize+1;
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer;
           move(line,pw^[i],xsize*2);
         end;
    1: begin
           pb:=bufferpointer;
           move(line,pb^[i],xsize);
      end;
    4: begin
           plong:=bufferpointer;
           move(line,plong^[i],xsize*4);
      end;
    end;
  end;
end;

procedure Timgbuffer.Getcolumn(x:integer; var col);
var
  pb,outb: ^byteArray;
  pw,outw: ^wordArray;
  plong,outlong: ^longArray;
  i,k : longint;
begin
  if xsize<=0 then exit
  else
  begin
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer; outw:=@col; k:=x;
           for i:=1 to ysize do
           begin outw^[i]:=pw^[k]; inc(k,xsize); end;
         end;
    1: begin
           pb:=bufferpointer; outb:=@col; k:=x;
           for i:=1 to ysize do
           begin outb^[i]:=pb^[k]; inc(k,xsize); end;
      end;
    4: begin
           plong:=bufferpointer; outlong:=@col; k:=x;
           for i:=1 to ysize do
           begin outlong^[i]:=plong^[k]; inc(k,xsize); end;
      end;
    end;
  end;

end;

procedure Timgbuffer.Putcolumn(var col; x:integer);
var
  pb,inb: ^byteArray;
  pw,inw: ^wordArray;
  plong,inlong: ^longArray;
  i,k : longint;
begin
  if xsize=0 then exit
  else
  begin
    case bufferbyte of
    0,2: begin
           pw:=bufferpointer; inw:=@col; k:=x;
           for i:= 1 to ysize do
           begin pw^[k]:=inw^[i]; inc(k,xsize); end;
         end;
    1: begin
           pb:=bufferpointer; inb:=@col; k:=x;
           for i:= 1 to ysize do
           begin pb^[k]:=inb^[i]; inc(k,xsize); end;
      end;
    4: begin
           plong:=bufferpointer; inlong:=@col; k:=x;
           for i:= 1 to ysize do
           begin plong^[k]:=inlong^[i]; inc(k,xsize); end;
      end;
    end;
  end;
end;

procedure swap4(var inp: integer);
var
  i:array[0..1] of word absolute inp;
  a,b:word;
begin
  a:=i[0]; b:=i[1];
  i[0]:=swap(b); i[1]:=swap(a);
end;

procedure Timgbuffer.swapbytes;
var
  pw: ^wordarray;
  plong: ^longArray;
  i: integer;
  m,w: word;
begin
  if (bufferbyte=0) then bufferbyte:=2;
  if (bufferbyte<2) then exit;
  case bufferbyte of
  2: begin
       pw:=bufferpointer; m:=0;
       for i:=1 to xsize*ysize do
       begin pw^[i]:=swap(pw^[i]); m:=max(m,pw^[i]); end;
       buffermax:=m;
     end;
  4: begin
       plong:=bufferpointer; m:=0;
       for i:=1 to xsize*ysize do
       begin
         swap4(plong^[i]);
         m:=max(m,plong^[i]);
       end;
       buffermax:=m;
     end;
  end;
end;

procedure Timgbuffer.flipvertical;
var
  pb1,pb2: ^byteArray;
  y : longint;
begin
  if (bufferbyte=0) then bufferbyte:=2;
  getmem(pb1,xsize*bufferbyte);
  getmem(pb2,xsize*bufferbyte);
  for y:=1 to ysize div 2 do
  begin
    Getline(y,pb1^);
    Getline(ysize-y+1,pb2^);
    Putline(pb1^,ysize-y+1);
    Putline(pb2^,y);
  end;
  freemem(pb1); freemem(pb2);
end;

procedure Timgbuffer.fliphorizontal;
var
  pb1,pb2: ^byteArray;
  x : longint;
begin
  if (bufferbyte=0) then bufferbyte:=2;
  getmem(pb1,xsize*bufferbyte);
  getmem(pb2,xsize*bufferbyte);
  for x:=1 to xsize div 2 do
  begin
    Getcolumn(x,pb1^);
    Getcolumn(xsize-x+1,pb2^);
    Putcolumn(pb1^,xsize-x+1);
    Putcolumn(pb2^,x);
  end;
  freemem(pb1); freemem(pb2);
end;

procedure interpolateLineb(var line1,line2:byteArray; n: integer);
var
  x,m:integer;
begin
  m:=line1[1]; m:=(m shl 1)+m;
  line2[1]:=line1[1];
  line2[2]:=(m+line1[2]) shr 2;
  for x:=2 to n-1 do
  begin
    m:=line1[x]; m:=(m shl 1)+m;
    line2[(x shl 1)-1]:=(m+line1[x-1]) shr 2;
    line2[x shl 1]:=(m+line1[x+1]) shr 2;
  end;
  m:=line1[n]; m:=(m shl 1)+m;
  line2[n shl 1]:=line1[n];
  line2[(n shl 1)-1]:=(m+line1[n-1]) shr 2;
end;

procedure interpolateLinew(var line1,line2:wordArray; n: integer);
var
  x,m:integer;
begin
  m:=line1[1]; m:=(m shl 1)+m;
  line2[1]:=line1[1];
  line2[2]:=(m+line1[2]) shr 2;
  for x:=2 to n-1 do
  begin
    m:=line1[x]; m:=(m shl 1)+m;
    line2[(x shl 1)-1]:=(m+line1[x-1]) shr 2;
    line2[x shl 1]:=(m+line1[x+1]) shr 2;
  end;
  m:=line1[n]; m:=(m shl 1)+m;
  line2[n shl 1]:=line1[n];
  line2[(n shl 1)-1]:=(m+line1[n-1]) shr 2;
end;

procedure Timgbuffer.gettemp(var imgtemp:Timgbuffer);
begin
  self.xsize:=imgtemp.xsize;
  self.ysize:=imgtemp.ysize;
  self.bufferbyte:=imgtemp.bufferbyte;
  self.yori:=imgtemp.yori;
  self.buffermax:=imgtemp.buffermax;
  self.buffermin:=imgtemp.buffermin;
  self.scalefactor:=imgtemp.scalefactor;
  self.bufferpointer:=imgtemp.bufferpointer;
  imgtemp.bufferpointer:=nil;
end;

procedure Timgbuffer.interpolate;
{bilinear interpolation: Matrixsize m x n -> 2m x 2n}
var
  inb,outb: ^byteArray;
  inw,outw: ^wordArray;
  p1,p2: pointer;
  imtemp: Timgbuffer;
  x,y,linesize: longint;
begin
  linesize:=2*max(xsize,ysize)*bufferbyte;
  getmem(p1,linesize); {linebuffer for input}
  getmem(p2,linesize);  {linebuffer for output}
  inw:=p1; outw:=p2; inb:=p1; outb:=p2;
  imtemp:=Timgbuffer.create;
  imtemp.init(2*xsize,2*ysize,bufferbyte);
  for y:=1 to ysize do
  begin
    self.Getline(y,inb^);
    case bufferbyte of
    1: interpolateLineb(inb^,outb^,self.xsize);
    0,2: interpolateLinew(inw^,outw^,self.xsize);
    end;
    imtemp.Putline(outb^,y);
  end;
  for x:=1 to imtemp.xsize do
  begin
    imtemp.GetColumn(x,inb^);
    case bufferbyte of
    1: interpolateLineb(inb^,outb^,self.ysize);
    0,2: interpolateLinew(inw^,outw^,self.ysize);
    end;
    imtemp.PutColumn(outb^,x);
  end;
  imtemp.buffermax:=imtemp.getmaxpixel;
  self.gettemp(imtemp);
  imtemp.free;
  freemem(p1);
  freemem(p2);
end;

function Timgbuffer.Duplicate: Timgbuffer;
var
  outbuffer: Timgbuffer;
  p: pointer;
  nbytes: integer;
  inb,outb: ^byteArray;
begin
  outbuffer:=Timgbuffer.create;
  outbuffer.xsize := xsize;
  outbuffer.ysize := ysize;
  outbuffer.yori := yori;
  outbuffer.bufferbyte := bufferbyte;
  outbuffer.bufferMax := bufferMax;
  outbuffer.ScaleFactor := ScaleFactor;
  if bufferbyte=0 then nbytes:=xsize*ysize*2
             else nbytes:=xsize*ysize*bufferbyte;
  getmem(p,nbytes);
  outbuffer.bufferpointer:=p;
  inb:=bufferpointer; outb:=p;
  move(inb^,outb^,nbytes);
  Duplicate:=outbuffer;
end;

function Timgbuffer.AddImage(img:Timgbuffer):boolean;
var
  pb1,pb2: ^byteArray;
  pw1,pw2: ^wordArray;
  plong1,plong2: ^longArray;
  //p1,p2: pointer;
  x,n,m: longint;
begin
  AddImage:=false;
  if ((xsize<>img.xsize) or (ysize<>img.ysize)) then exit;
  pb1:=bufferpointer; pb2:=img.bufferpointer;
  pw1:=bufferpointer; pw2:=img.bufferpointer;
  plong1:=bufferpointer; plong2:=img.bufferpointer;
  n:=xsize*ysize;
  m:=0;
  for x:=1 to n do
  begin
    case bufferbyte of
    0,2: begin
           case img.bufferbyte of
           0,2: m:=pw1^[x]+pw2^[x];
           1:   m:=pw1^[x]+pb2^[x];
           4:   m:=pw1^[x]+plong2^[x];
           end;
           pw1^[x]:=min(m,65535);
           buffermax:=max(pw1^[x],buffermax);
         end;
    1:   begin
           case img.bufferbyte of
           0,2: m:=pb1^[x]+pw2^[x];
           1:   m:=pb1^[x]+pb2^[x];
           4:   m:=pb1^[x]+plong2^[x];
           end;
           pb1^[x]:=min(m,255);
           buffermax:=max(pb1^[x],buffermax);
         end;
    4:   begin
           case img.bufferbyte of
           0,2: m:=plong1^[x]+pw2^[x];
           1:   m:=plong1^[x]+pb2^[x];
           4:   m:=plong1^[x]+plong2^[x];
           end;
           plong1^[x]:=m;
           buffermax:=max(plong1^[x],buffermax);
         end;
    end;
  end;
  AddImage:=true;
end;

procedure Timgbuffer.ByteToWord; // changes byte matrix to word matrix
var
  inb: ^byteArray;
  outw: ^wordArray;
  i: integer;
begin
  if (bufferbyte<>1) then exit;
  getmem(outw,xsize*ysize*2);
  inb:=bufferpointer;
  for i:=1 to xsize*ysize do outw^[i]:=inb^[i];
  bufferbyte:=2;
  freemem(bufferpointer);
  bufferpointer:=outw;
end;

procedure Timgbuffer.IntToWord(offset:integer);
  //changes signed (smallint) to word matrix:
  //offset is added to pixel value, resulting value is truncated to word range
var
  inb: ^intArray;
  outw: ^wordArray;
  i: integer;
begin
  inb:=bufferpointer;
  outw:=bufferpointer;
  for i:=1 to xsize*ysize do outw^[i]:=max(0,min(65535,offset+ inb^[i]));
end;

procedure Timgbuffer.WordtoInt(offset:integer);
  //changes word to signed (smallint) matrix:
  //offset is subtracted from pixel value, resulting value is truncated to int range
var
  outb: ^intArray;
  inw: ^wordArray;
  i: integer;
begin
  inw:=bufferpointer;
  outb:=bufferpointer;
  for i:=1 to xsize*ysize do outb^[i]:=max(-32768,min(32767,inw^[i]-offset));
end;

procedure Timgbuffer.FitsToLongint(offset:double);
  //subtracts fits offset from signed (longint) matrix:
const
  ofs32=#$7FFFFFFF;
var
  outlong: ^LongArray;
  inw: ^IntArray;
  inlong: ^LongArray;
  i,ofs2: integer;
  useOfs32: boolean;
begin
  if offset<=2147483648.0 then ofs2:=round(offset-1) else exit;
  inw:=bufferpointer; inlong:=bufferpointer;
  outlong:=bufferpointer;
  case bufferbyte of
  2: {not implemented, buffer size must be changed here (shortint to longint)};
  4: begin
          for i:=1 to xsize*ysize do outlong^[i]:=inlong^[i] + ofs2 +1;
     end;
  end;
end;

procedure Timgbuffer.UpdateFitsHeader;
begin
     setfitskey(fitsheader,'NAXIS1',inttostr(xsize));
     setfitskey(fitsheader,'NAXIS2',inttostr(ysize));
     //setfitskey(fitsheader,'','');
     //setfitskey(fitsheader,'','');
end;

function Timgbuffer.Readfitsfile(fitsfilename:string):boolean;
var
  line: Tfitsline;
  key: Tfitskey;
  param: string;
  f: file;
  i,k: integer;
  complete: boolean;
  rparam: double;
  iparam: integer;
  bparam: boolean;
  sparam:string;
  err1,err2: integer;
  day,month,year,h,m: integer;
  sec,hour: single;
begin
  readfitsfile:=false;
  complete:=false;
  filename:=fitsfilename;
  if lookup(f,filename)<>0 then exit;
  if iopen(f,byteread)<>0 then exit;
  k:=0; imageblock:=0; i:=0;
  while ((not complete) and (k<=3)) do
  begin
   if IREADbyte(2880,fitsheader[k*36+1],k*2880,f)<2880
     then exit;
   i:=1;
   repeat
    line:=fitsheader[k*36+i];
    headerlines:=k*36+i;
    key:=copy(line,1,8);
    param:=copy(line,11,70);
    sparam:=''; rparam:=0.0; iparam:=0;
    if pos(chr(39),param)>0 then
      sparam:=copy(param,pos(chr(39),param)+1,70-pos(chr(39),param));
    if pos(chr(39),sparam)>0 then
      sparam:=copy(sparam,1,pos(chr(39),sparam)-1);
    param:=copy(line,11,20);
    if pos('/',param)>0 then param:=copy(param,1,pos('/',param)-1);
    while (pos(' ',param)>0) do delete(param,pos(' ',param),1);
    iparam:=0; rparam:=0.0;
    if (line[30]='T') then bparam:=true else bparam:=false;
    val(param,iparam,err1);
    val(param,rparam,err2);
    if (((i=1) and (k=0)) and (key<>'SIMPLE  ')) then
    begin closefile(f); exit; end
    else if (key='END     ') then complete:=true
    else if (key='NAXIS1  ') then
    begin
      if (err1=0) then xsize:=iparam;
    end else if (key='NAXIS2  ') then
    begin
      if (err1=0) then ysize:=iparam;
    end else if (key='BITPIX  ') then
    begin
      case iparam of
      8: bufferbyte:=1;
      16: bufferbyte:=2;
      32: bufferbyte:=4;
      else exit;
      end;
    end else if (key='BZERO   ') then
    begin
      Offset1:=Rparam;
    end else if (key='BSCALE  ') then
    begin
      Scalefactor:=Rparam;
    end else if (key='DATE-OBS') then
    begin
      convertdate(sparam,year,month,day);
    end else if ((key='UT-START') or (key='TIME-OBS')) then
    begin
      converttime(sparam,h,m,sec);
    end;
    inc(i);
   until ((key='END     ') or (i>36));
   inc(k);
   imageblock:=k*2880;
  end;//while not complete
  if ((xsize>0) and (ysize>0)) then
  begin
    init(xsize,ysize,bufferbyte);
  end;
  JDstart:=JD(day,month,year,h+m/60+sec/3600.0);
  DateTime:= Inttostr(year)+'-'+Inttostr(month)+'-'+Inttostr(day)
             +' '+Inttostr(h)+':'+Inttostr(m);
  if IREADbyte(xsize*ysize*bufferbyte,bufferpointer^,imageblock,f)
     <xsize*ysize*bufferbyte then
     begin closefile(f);exit; end;
  if (bufferbyte=2) or (bufferbyte=4) then Swapbytes;
  Wordmode:=false;
  closefile(f);
  if (bufferbyte=2) and (Offset1=32768.0) then
  begin
    IntToword(round(offset1));
    Offset1:=0;
    wordmode:=true;
  end else if (bufferbyte=4) and (Offset1<>0) then
  begin
    FitstoLongInt(Offset1);
  end;
  GetMinmax;
  FlipVertical;
  readfitsfile:=true;
end;

function Timgbuffer.ReadBMPfile(BMPfilename:string):boolean;
var
   pb:^byteArray;
   x,y:integer;
   bmp:Tbitmap;
   day,month,year,hour,minute,second,millisecond: word;
   fdate: Tdatetime;
begin
     bmp:=Tbitmap.Create;
     ReadBMPfile:=false;
     try
        bmp.LoadFromFile(BMPfilename);
        self.init(bmp.width,bmp.height,2);
        self.filename:=BMPfilename;
        for y:=0 to bmp.height-1 do
        begin
          pb:=bmp.ScanLine[y];
          for x:=1 to bmp.width do self.putpixel(x,y+1,pb^[1+3*(x-1)]);
        end;
     finally
        bmp.free;
     end;
     fdate:=FileDateToDateTime(FileAge(BMPFileName));
     //DecodeDateTime(fdate,Year, Month, Day, Hour, Minute, Second, MilliSecond);
     DecodeDate(fdate,Year,Month,DAY);
     DecodeTime(fdate,hour,minute,second,Millisecond);
     JDstart:=JD(day,month,year,hour+minute/60+second/3600.0+MilliSecond/3600000);
     DateTime:= Inttostr(year)+'-'+Inttostr(month)+'-'+Inttostr(day)
             +' '+Inttostr(hour)+':'+Inttostr(minute)+':'+inttostr(Second);
     GetMinmax;
     FlipVertical;
     ReadBMPfile:=true;
end;

function setFitskey(var header:Tfitsheader; key:Tfitskey; value:string):boolean;
var
  i,k:integer;
  line: Tfitsline;
  ikey: Tfitskey;
  iend: integer;
begin
  i:=0; setfitskey:=false;
  iend:=-1;
  repeat
    inc(i); line:=header[i];
    ikey:=copy(line,1,8);
    while (length(key)<8) do key:=key+' ';
    if ikey='END     ' then iend:=i;
  until ((ikey=key) or (i=maxheaderlines) or (ikey='END     '));
  if (ikey=key) then
  begin
    line[9]:='='; for k:=10 to 80 do line[k]:=' ';
    for k:=1 to length(value) do line[10+k]:=value[k];
    header[i]:=line;
    setfitskey:=true;
  end else
  begin
    if (iend>0) and (iend<maxheaderlines) then
    begin
         header[iend+1]:=header[iend];
         line:=header[iend];
         for k:=1 to 8 do line[k]:=key[k];
         line[9]:='='; for k:=10 to 80 do line[k]:=' ';
         for k:=1 to length(value) do line[10+k]:=value[k];
         header[iend]:=line;
         setfitskey:=true;
    end else
    begin //init
         for k:=1 to 8 do line[k]:=key[k];
         line[9]:='='; for k:=10 to 80 do line[k]:=' ';
         for k:=1 to length(value) do line[10+k]:=value[k];
         header[1]:=line;
         ikey:='END     ';
         for k:=1 to 8 do line[k]:=ikey[k];
         for k:=9 to 80 do line[k]:=' ';
         header[2]:=line;
         setfitskey:=true;
    end;
  end;
end;

function Timgbuffer.Writefitsfile(fitsfilename:string):boolean;
var
  i,k:integer;
  f: file;
begin
  Writefitsfile:=false;
  UpdateFitsHeader;//21.12.03
  assignfile(f,fitsfilename);
  {$I+}rewrite(f,1);{$I-}
  if IOresult<>0 then exit;
  FlipVertical;
  WordtoInt(32768);
  swapbytes;
  //Setfitskey(fitsheader,'BZERO',floattostr(offset1));
  for k:= 0 to ((headerlines-1) div 36) do
  begin
   if Iwritbyte(2880,fitsheader[k*36+1],k*2880,f)<2880
     then
     begin closefile(f);exit;end;
  end;
  k:=(headerlines-1) div 36;
  if Iwritbyte(xsize*ysize*bufferbyte,bufferpointer^,(k+1)*2880,f)
     <xsize*ysize*bufferbyte then
     begin closefile(f);exit; end;
  closefile(f);
  Writefitsfile:=true;
end;//writefitsfile



{IMAGE CACHE}
{-----------}
constructor Timagecache.create;
begin
  inherited create;
  lastindex:=0;
  memcount:=0;
  lastID:=0;
  maxindex:=512;
  maxmem:=4000000;
end;

destructor Timagecache.destroy;
var
  i: integer;
begin
  for i:=1 to lastindex do
  begin
    cbuffer[i].free; cbuffer[i]:=nil;
  end;
  inherited destroy;
end;

function TimageCache.getFreeBuffer:integer;
{frees 1 buffer, if possible, and returns free buffer number,
 or 0, if no buffer available}
var
i,mp,index,nbytes,freeindex: integer;
begin
  if lastindex<1 then getFreeBuffer:=0
  else begin
    mp:=lastID+10000;
    index:=0; freeindex:=0;
    for i:=1 to lastindex do
    begin
      if (ID[i]>0)then
      begin
        if ((ID[i]+accesscount[i])<mp) then
        begin
          mp:=ID[i]+accesscount[i];
          index:=i;
        end;
      end else freeindex:=i;
    end;
    if index=0 then index:=freeindex;
    if ((index>0) and (ID[index]>0)) then
    begin
      ID[index]:=0; accesscount[index]:=0;
      with cbuffer[index] do nbytes:=xsize*ysize*bufferbyte;
      cbuffer[index].free; cbuffer[index]:=nil;
      memcount:=memcount-nbytes;
    end;
    getfreebuffer:=index;
  end;
end;

function TimageCache.PutImagetoCache(img:Timgbuffer):TcacheID;
var
  newindex,newmem: integer;
  nbytes: integer;
  newID: TcacheID;
begin
  PutImagetoCache:=-1;
  if (img=nil) then exit;
  newindex:=lastindex+1;
  newID:=lastID+1;
  nbytes:=img.xsize*img.ysize*img.bufferbyte;
  if (nbytes>maxmem) then exit; //modif. (V.1.28) 11.2.98
  newmem:=memcount+nbytes;
  if ((newindex>maxindex) or (newmem>maxmem)) then
  begin
    repeat
      newindex:=getfreeBuffer; newmem:=memcount+nbytes;
    until ((newindex<=0) or (newmem<=maxmem));     //modif. (V.1.28) 11.2.98
    if newindex<=0 then exit;
  end else lastIndex:=newIndex;
  cbuffer[newindex]:=img.duplicate;
  ID[newindex]:=newID;
  accesscount[newindex]:=0;
  lastID:=newID;
  memcount:=newmem;
  PutImagetoCache:=newID;
end;

function TimageCache.loadfromCache(cacheID:TcacheID; var img:Timgbuffer):integer;
var
  i: integer;
begin
  loadfromCache:=-1;
  if cacheID<=0 then exit;
  for i:=1 to lastindex do
  begin
    if (ID[i]=cacheID) then
    begin
      img.free;// Bugfix V. 1.13
      img:=cbuffer[i].duplicate;
      loadfromCache:=ID[i];
      inc(accesscount[i]);
      exit;
    end;
  end;
end;

function TimageCache.DisposeCache(cacheID:TcacheID):integer;
var
  i: integer;
  nbytes: longint;
begin
  DisposeCache:=0;
  if cacheID<=0 then exit;
  for i:=1 to lastindex do
  begin
    if (ID[i]=cacheID) then
    begin
      ID[i]:=0;
      with cbuffer[i] do nbytes:=xsize*ysize*bufferbyte;
      cbuffer[i].free; cbuffer[i]:=nil;
      memcount:=memcount-nbytes;   //bugfix (v.1.28) 11.2.98
      exit;
    end;
  end;
end;

{ BITMAPS }
{---------}
procedure TBitmapRGB.LoadfromArray(arr:pointer; xsize,ysize:Integer);
var
    bmpinfo:TBitmapinfo;
Begin
    with bmpinfo.bmiheader Do
    Begin

          biHeight := ysize; biWidth := xsize;
          biSize := SizeOf(bmpInfo.bmiHeader);
          biPlanes := 1; biBitCount := 24;
          biCompression:=0;
          biSizeImage:=0;
          biXPelsPerMeter:=0;
          biYPelsPerMeter:=0;
          biClrUsed:=0;
          biClrImportant:=0;

     End;

     //DeleteObject(inherited handle);
     inherited handle:=CreateDIBitmap(getdc(0), bmpinfo.bmiheader,
                           cbm_Init, arr, bmpinfo, dib_RGB_Colors);

End;

destructor Tbitmap256.destroy;
begin
  DeleteObject(Handle);
  ReleaseHandle;
  inherited destroy;
end;

(*
function TBitmap256.LoadfromArray(arr:pointer;xsize,ysize:Integer;HPal:Hpalette):boolean;
// V.1.35
type
  tpalentries = array[0..255] of tpaletteentry;
const
 ncolors=256;
var
    bmpinfo256:^TBitmapinfo256;
    bmpinfo,bmpinfo2: ^Tbitmapinfo;
    palcolors: ^tpalentries;
    p1,p2: pointer;
    i,n: integer;
    holdpal: hpalette;
Begin
    getmem(p1,sizeof(TBitmapinfo256));
    getmem(p2,sizeof(TBitmapinfo256));
    getmem(palcolors,sizeof(tpalentries));
    bmpinfo:=p1; bmpinfo256:=p2; bmpinfo2:=p2;
    fillchar(bmpinfo256^,sizeof(Tbitmapinfo256),#0); //initialize
    with bmpinfo^.bmiheader Do
    Begin
       biHeight := ysize; biWidth := xsize;
       biSize := SizeOf(bmpInfo.bmiHeader);
       biPlanes := 1; biBitCount := 8;
       biCompression:=BI_RGB;
       biSizeImage:=0;
       biXPelsPerMeter:=0;
       biYPelsPerMeter:=0;
       biClrUsed:=256;
       biClrImportant:=256;
     End;
     Move(bmpinfo^,bmpinfo256^,sizeOf(Tbitmapinfo));
     n:=GetPaletteEntries(hpal,0,ncolors,palcolors^);

     for i:=0 to (ncolors-1) do
     begin
       with bmpinfo256^.bmicolors[i] do
       begin
         rgbred:=palcolors^[i].pered;
         rgbgreen:=palcolors^[i].pegreen;
         rgbblue:=palcolors^[i].peblue;
         rgbreserved:=0;
       end;
     end;

     ignorePalette:=false;
     //DeleteObject(Inherited Palette);
     Inherited Palette:=HPal;
     releaseDC(0,canvas.handle);
     canvas.handle:=getdc(0);
     holdpal:=selectPalette(canvas.handle,HPal,false);
     RealizePalette(canvas.handle);
     monochrome:=false;
     //DeleteObject(inherited handle);
     inherited handle:= CreateDIBitmap(canvas.handle, bmpinfo^.bmiheader,
                        cbm_Init, arr, bmpinfo2^, dib_rgb_Colors);
     if (handle<>0) then LoadfromArray:=true
     else begin
       LoadfromArray:=false;
       {ShowMessage('createDIbitmap error');}
     end;
     freemem(p1);freemem(p2);
     freemem(palcolors);
  End;
*)
function TBitmap256.LoadfromArray(arr:pointer;xsize,ysize:Integer;HPal:Hpalette):boolean;
//V. 1.37
type
  tpalentries = array[0..255] of tpaletteentry;
const
 ncolors=256;
var
    bmpinfo256:^TBitmapinfo256;
    bmpinfo,bmpinfo2: ^Tbitmapinfo;
    palcolors: ^tpalentries;
    p1,p2: pointer;
    i,n: integer;
    holdpal: hpalette;
    DC: hDC;
Begin
    getmem(p1,sizeof(TBitmapinfo256));
    getmem(p2,sizeof(TBitmapinfo256));
    getmem(palcolors,sizeof(tpalentries));
    bmpinfo:=p1; bmpinfo256:=p2; bmpinfo2:=p2;
    fillchar(bmpinfo256^,sizeof(Tbitmapinfo256),#0); //initialize
    with bmpinfo^.bmiheader Do
    Begin
       biHeight := ysize; biWidth := xsize;
       biSize := SizeOf(bmpInfo.bmiHeader);
       biPlanes := 1; biBitCount := 8;
       biCompression:=BI_RGB;
       biSizeImage:=0;
       biXPelsPerMeter:=0;
       biYPelsPerMeter:=0;
       biClrUsed:=256;
       biClrImportant:=256;
     End;
     Move(bmpinfo^,bmpinfo256^,sizeOf(Tbitmapinfo));
     n:=GetPaletteEntries(hpal,0,ncolors,palcolors^);

     for i:=0 to (ncolors-1) do
     begin
       with bmpinfo256^.bmicolors[i] do
       begin
         rgbred:=palcolors^[i].pered;
         rgbgreen:=palcolors^[i].pegreen;
         rgbblue:=palcolors^[i].peblue;
         rgbreserved:=0;
       end;
     end;

     ignorePalette:=false;
     //DeleteObject(Inherited Palette);
     Inherited Palette:=HPal;
     releaseDC(0,canvas.handle);
     canvas.handle:=getdc(0);
     selectPalette(canvas.handle,HPal,false);
     RealizePalette(canvas.handle);
     monochrome:=false;
     //DeleteObject(inherited handle);
     ReleaseHandle;
     //DC:=getDC(0);
     inherited handle:= CreateDIBitmap({DC}canvas.handle, bmpinfo^.bmiheader,
                        cbm_Init, arr, bmpinfo2^, dib_rgb_Colors);
     if (handle<>0) then LoadfromArray:=true
     else begin
       LoadfromArray:=false; ErrorReporting:=true;
       ShowMessage('createDIbitmap error'); ErrorReporting:=false;
     end;
     //ReleaseDC(0,DC);
     freemem(p1);freemem(p2);
     freemem(palcolors);
  End;


  procedure Tbitmap256.LoadfromBuffer( img: TImgBuffer; ilt,iut:integer;HPal:Hpalette);
  var
    pb: ^byteArray;
    pba: ^byteArray;
    pwa: ^wordArray;
    pla: ^longArray;
    i,d: integer;
    c: single;
    bmpmax: byte;
    xdim,ydim,nfill,mfill: integer;
    pa: pointer; //1.42
    hpixelbuffer: Thandle; //1.42

  begin
    case dcnumcolors of
    256: bmpmax:=127;
    else bmpmax:=255;
    end;
    xdim:=img.xsize; ydim:=img.ysize;
    nfill:=3-((xdim+3) mod 4);
    if nfill<>0 then xdim:=xdim+nfill;
    //getmem(pb,xdim*ydim);
    hpixelbuffer:=GlobalAlloc(GHND,xdim*ydim);//1.42
    pb:=GlobalLock(hPixelBuffer); //1.42

    pba:=img.bufferpointer; pwa:=img.bufferpointer;
    pla:=img.bufferpointer;
    if img.wordmode then
    begin
      if ilt<0 then ilt:=round(0.01*img.buffermax*(-ilt));
      if iut<0 then iut:=round(0.01*img.buffermax*(-iut));
    end;
    if iut<=ilt then iut:=ilt+1;
    if nfill=0 then
    begin
     case img.bufferbyte of
     1: begin
         if (img.buffermax-ilt>1) then c:=bmpmax/(iut-ilt)
                                  else c:=0;
         for i:=1 to img.xsize*img.ysize do
         begin
           d:=pba^[i]-ilt;
           if d>0 then pb^[i]:=min(round(c*d),bmpmax)
                  else pb^[i]:=0;
         end;
       end;
     0,2: begin
           if (img.buffermax-ilt>1) then c:=bmpmax/(iut-ilt)
                                    else c:=0;
           for i:=1 to img.xsize*img.ysize do
           begin
             d:=pwa^[i]-ilt;
             if d>0 then pb^[i]:=min(round(c*d),bmpmax)
                    else pb^[i]:=0;
           end;
         end;
     4: begin
           if (img.buffermax-ilt>1) then c:=bmpmax/(iut-ilt)
                                    else c:=0;
           for i:=1 to img.xsize*img.ysize do
           begin
             d:=pla^[i]-ilt;
             if d>0 then pb^[i]:=min(round(c*d),bmpmax)
                    else pb^[i]:=0;
           end;
         end;
     end;//case
    end else   //x-size of bitmap greater than that of image buffer
    begin
     case img.bufferbyte of
     1: begin
         if (img.buffermax-ilt>1) then c:=bmpmax/(iut-ilt)
                                  else c:=0;
         for i:=1 to img.xsize*img.ysize do
         begin
           mfill:= ((i-1) div img.xsize)*nfill;
           d:=pba^[i]-ilt;
           if d>0 then pb^[i+mfill]:=min(round(c*d),bmpmax)
                  else pb^[i+mfill]:=0;
         end;
       end;
     0,2: begin
           if (img.buffermax-ilt>1) then c:=bmpmax/(iut-ilt)
                                    else c:=0;
           for i:=1 to img.xsize*img.ysize do
           begin
             mfill:= ((i-1) div img.xsize)*nfill;
             d:=pwa^[i]-ilt;
             if d>0 then pb^[i+mfill]:=min(round(c*d),bmpmax)
                    else pb^[i+mfill]:=0;
           end;
         end;
     4: begin
           if (img.buffermax-ilt>1) then c:=bmpmax/(iut-ilt)
                                    else c:=0;
           for i:=1 to img.xsize*img.ysize do
           begin
             mfill:= ((i-1) div img.xsize)*nfill;
             d:=pla^[i]-ilt;
             if d>0 then pb^[i+mfill]:=min(round(c*d),bmpmax)
                    else pb^[i+mfill]:=0;
           end;
         end;
     end;//case

    end;
    pa:=pb; //1.42
    if (not LoadfromArray(pa,xdim,-ydim,Hpal)) then   //18.1.99, 1.42
    begin
    end;
    //freemem(pb);
    GlobalUnlock(hPixelBuffer); //1.42
    GlobalFree(hPixelBuffer);   //1.42
  end;

function CreatePalette256(pal256: TPalette256):HPalette;
var
  rpal: tlogpalette absolute pal256;
  k: integer;
  palsize: integer;
begin
  if (Getnumcolors=256) then palsize:=128
                      else palsize:=256;
  for k:=0 to palsize-1 do
  begin
    with pal256.palpalentry[k] do
         begin
           case palsize of
           128: begin
                  pered:=pal256.palpalentry[k shl 1].pered;
                  pegreen:=pal256.palpalentry[k shl 1].pegreen;
                  peblue:=pal256.palpalentry[k shl 1].peblue;
                end;
           256: begin
                end;
           end;
           peflags:=PC_NOCOLLAPSE;
         end;
  end;
  pal256.palnumentries:=palsize;
  CreatePalette256:=CreatePalette(rpal);
end;

function LoadPalette(p:integer):Hpalette; //obsolete
const
 ncolors=256;
var
    i,n: integer;
    logpal: ^Tlogpalette;
Begin
    getmem(logpal,sizeof(tlogpalette)+(ncolors-1)*sizeof(tpaletteentry));
{$R-}
    FOR I:=0 to (ncolors-1) do
    begin
       with logpal^.palpalentry[i] do
       begin
         pered:=i; pegreen:=i; peblue:=i;
         peflags:=PC_NOCOLLAPSE;
       end;
    end;
    logpal^.palversion:=$300;
    logpal^.palnumentries:=ncolors;
    LoadPalette:=CreatePalette(logpal^);

    //n:=GetPaletteEntries(ThePalette,0,256,logpal^.palpalentry);

    freemem(logpal);
end;

function GetnumColors: integer;
begin
  DCnumColors:=getdevicecaps(getdc(0),SIZEPALETTE);
  //DCnumcolors:=65000;
  GetnumColors:=DCnumcolors;
end;

procedure StartImageCache(cacheMB:integer);
//var
  //heapstatus: theapstatus;
begin
  //heapstatus:=GetHeapStatus;
  //ShowMessage('totalfree='+inttostr(heapstatus.TotalFree));
  if (cacheMB=0) then
  begin
    if (ImageCache1<>nil) then StopImageCache;
    exit;
  end;
  if (ImageCache1<>nil) then StopImageCache;
  Imagecache1:=TimageCache.create;
  if cacheMB>0 then ImageCache1.maxmem:=cacheMB*1048576;
  ImageCacheEnabled:=true;
end;

procedure StopImageCache;
begin
  ImageCacheEnabled:=false;
  ImageCache1.free;
end;

function GetColortableList(var CtableList,CtableFileList: Tstringlist):integer;
         // returns a list of color table names corresp. to a list
         // of filenames
var
  FSR : TSearchRec;
  i,j : Integer;
  S : string;
  ctfile: textfile;
begin
  GetColortableList:=0;
  if CtableList1=nil then
  begin
    CtableList1:=Tstringlist.create;
    CtableFileList1:=Tstringlist.create;
    CtableList1.Add('*Grayscale linear');
    CtableFileList1.add('');
    j:=findfirst(Progdir+'*.GVC', faArchive, FSR);
    while (j=0) do
    begin
      AssignFile(ctfile,Progdir+FSR.name);
      {$I-}reset(ctfile);
      readln(ctfile,S);S:=copy(S,1,20);
      if (stupcase(S)='GAMMAVIEW COLORTABLE') then
      begin
        readln(ctfile,S);
        CtableList1.Add(S);
        CtableFileList1.add(FSR.name);
      end;
      close(ctfile);
      j:=FindNext(FSR);
    end; {$I+}
  end;
  CtableList:=CtableList1; CtableFileList:=CtableFileList1;
  GetColortableList:=CtableList1.count;
end;

function GetColortable( filename:string; var colortable:Tpalette256;
                        inverted:boolean): boolean;
         // returns a 256 color palette from a colortable file
         // format of colortable file:
         // ASCII-file, first line: GammaView colortable 256 R,G,B
         // second line: <name of colortable>
         // line 3..257 contain the color values for index 0..255
         // format <red>,<green>,<blue>
         // !! memory for colortable must be provided by the calling procedure !!
var
  ctfile: textfile;
  i,k,err: integer;
  S,Sn,name: string;
  rgb: array[0..255] of record r,g,b:byte; end;
  rgbinput: array[1..3] of integer;
begin
  GetColortable:=false;
  if (filename='') then     { gray scale }
  begin
    for i:=0 to 255 do
    begin
      rgb[i].r:=i; rgb[i].g:=i; rgb[i].b:=i;
    end;
  end else
  begin
    assignfile(ctfile,progdir+filename);
    {$I-}reset(ctfile);{$I+}
    if (IOresult<>0) then exit;
    if (eof(ctfile)) then begin close(ctfile); exit; end;
    readln(ctfile,S); S:=stupcase(copy(S,1,20));
    if (S<>'GAMMAVIEW COLORTABLE') then exit;
    if (eof(ctfile)) then begin close(ctfile); exit; end;
    readln(ctfile,name);
    for i:=0 to 255 do
    begin
      if (eof(ctfile)) then begin close(ctfile); exit; end;
      readln(ctfile,S);
      for k:=1 to 3 do
      begin
        if (k<3) then
        begin
          Sn:=copy(S,1,pos(',',S)-1); S:=copy(S,pos(',',S)+1,Length(S));
        end else Sn:=S;
        val(Sn,rgbinput[k],err);
        if ((err<>0) or (rgbinput[k]<0) or (rgbinput[k]>255)) then
        begin close(ctfile); exit; end;
      end;
      rgb[i].r:=rgbinput[1]; rgb[i].g:=rgbinput[2]; rgb[i].b:=rgbinput[3];
    end;
    close(ctfile);
  end;
  colortable.palversion:=$300;
  colortable.palnumentries:=256;
  for i:=0 to 255 do
  begin
    with colortable.palpalentry[i] do
    begin
      if (not inverted) then
      begin
        pered:=rgb[i].r; pegreen:=rgb[i].g;
        peblue:=rgb[i].b; peflags:=PC_NOCOLLAPSE;
      end else
      begin
        pered:=rgb[255-i].r; pegreen:=rgb[255-i].g;
        peblue:=rgb[255-i].b; peflags:=PC_NOCOLLAPSE;
      end;
    end;
  end;
  GetColortable:=true;
end;

function GetGrayPalette(var gPal:TPalette256; inverted: boolean ):boolean;
begin
  GetGrayPalette:=GetColortable('',gpal,inverted);
end;

begin
  DC0:=getDC(0);
  GetnumColors;
  //StartImageCache;
  //GetColortableList(CtableList1,CtableFileList1);
end.
